import requests
from jose import jwt

from six.moves.urllib_parse import urlencode

from ._property import overwritable_property
from ._request_auth import HTTPBasicAuth, HTTPBearerAuth


class TokenResponse(object):
	def __init__(self, data, client=None):
		self._data = data
		self._client = client

	@property
	def access_token(self):
		return self._data["access_token"]

	@property
	def id_token(self):
		return self._data["id_token"]

	@property
	def userinfo(self):
		try:
			return self._userinfo
		except AttributeError:
			userinfo = self._client.get_userinfo(self.access_token)
			if self.id["sub"] != userinfo["sub"]:
				return None
			self._userinfo = userinfo
			return self._userinfo

class OpenIDClient(object):
	def __init__(self, url, client_id=None, client_secret=None, initial_auth=None, initial_access_token=None, registration_client_uri=None, registration_auth=None, registration_access_token=None):
		self.url = url
		self.client_id = client_id
		self.client_secret = client_secret

		if initial_auth:
			self.initial_auth = initial_auth
		self.initial_access_token = initial_access_token

		self.registration_client_uri = registration_client_uri
		if registration_auth:
			self.registration_auth = registration_auth
		self.registration_access_token = registration_access_token

		self._configuration = self.get_configuration()

	@overwritable_property
	def auth(self):
		if self.client_secret:
			return HTTPBasicAuth(self.client_id, self.client_secret)
		else:
			return None

	def get_configuration(self):
		r = requests.get(self.url + ".well-known/openid-configuration")
		r.raise_for_status()
		return r.json()

	def translate_scope_in(self, scope):
		return scope

	def translate_scope_out(self, scope):
		return scope

	def translate_userinfo(self, userinfo):
		return userinfo

	def get_id(self, token_response):
		return jwt.decode(
			token_response.id_token,
			self.keys,
			algorithms = ['RS256'],
			audience = self.client_id,
			issuer = self.issuer,
			options = dict(verify_at_hash = False),
		)

	@property
	def issuer(self):
		return self._configuration["issuer"]

	@property
	def authorization_endpoint(self):
		return self._configuration["authorization_endpoint"]

	@property
	def token_endpoint(self):
		return self._configuration["token_endpoint"]

	@property
	def jwks_uri(self):
		return self._configuration["jwks_uri"]

	@property
	def userinfo_endpoint(self):
		return self._configuration["userinfo_endpoint"]

	@property
	def end_session_endpoint(self):
		return self._configuration.get("end_session_endpoint")

	@property
	def registration_endpoint(self):
		return self._configuration.get("registration_endpoint")

	@property
	def keys(self):
		r = requests.get(self._configuration["jwks_uri"])
		r.raise_for_status()
		return r.content.decode('utf-8')

	def get_userinfo(self, access_token):
		r = requests.get(self.userinfo_endpoint, headers=dict(
			Authorization = "Bearer " + access_token,
		))
		r.raise_for_status()
		data = r.json()
		return self.translate_userinfo(data)

	def authorize(self, redirect_uri, state='', scope=('openid',)):
		scope = set(self.translate_scope_in(scope))
		return self.authorization_endpoint + "?" + urlencode(dict(
			client_id=self.client_id,
			response_type="code",
			redirect_uri=redirect_uri,
			state=state,
			scope=" ".join(scope),
		))

	def request_token(self, redirect_uri, code):
		r = requests.post(self.token_endpoint, auth=self.auth, data=dict(
			grant_type="authorization_code",
			redirect_uri=redirect_uri,
			code=code,
		), headers={'Accept': 'application/json'})
		r.raise_for_status()
		resp = TokenResponse(r.json(), self)

		if "scope" in resp._data:
			resp.scope = set(self.translate_scope_out(set(resp._data["scope"].split(" "))))
		if not hasattr(resp, "scope") or "openid" in resp.scope:
			resp.id = self.get_id(resp)

		return resp

	def end_session(self, post_logout_redirect_uri, state, id_token_hint=''):
		if not self.end_session_endpoint:
			raise NotImplementedError("This OP does not support RP-initiated logout.")

		return self.end_session_endpoint + "?" + urlencode(dict(
			client_id=client_id, # See https://bitbucket.org/openid/connect/issues/914/session-5-missing-client_id-parameter
			post_logout_redirect_uri=post_logout_redirect_uri,
			state=state,
			id_token_hint=id_token_hint,
		))

	@overwritable_property
	def initial_auth(self):
		if self.initial_access_token:
			return HTTPBearerAuth(self.initial_access_token)
		else:
			# TODO? Write an OAuth extension for using Basic auth for registration_endpoint.
			return self.auth

	def register(self, auth=None, access_token=None, **client_config):
		if not self.registration_endpoint:
			raise NotImplementedError("This OP does not support Dynamic Client Registration.")

		if not auth and access_token:
			auth = HTTPBearerAuth(access_token)

		if not auth:
			auth = self.initial_auth

		r = requests.post(self.registration_endpoint, auth=auth, json=client_config)
		r.raise_for_status()
		data = r.json()
		client = type(self)(self.url, client_id=data['client_id'], client_secret=data.get('client_secret'), registration_client_uri=data.get('registration_client_uri'), registration_access_token=data.get('registration_access_token'))
		return client, data

	@overwritable_property
	def registration_auth(self):
		if self.registration_access_token:
			return HTTPBearerAuth(self.registration_access_token)
		else:
			# TODO Write an OAuth extension for using normal auth for registration_client_uri.
			return self.auth

	@property
	def client(self):
		if not self.registration_client_uri:
			raise NotImplementedError("registration_client_uri was not provided.")

		r = requests.get(self.registration_client_uri, auth=self.registration_auth)
		r.raise_for_status()
		return r.json()

	@client.setter
	def client(self, config):
		if not self.registration_client_uri:
			raise NotImplementedError("registration_client_uri was not provided.")

		r = requests.put(self.registration_client_uri, auth=self.registration_auth, json=config)
		r.raise_for_status()
		return r.json()

	@client.deleter
	def client(self):
		if not self.registration_client_uri:
			raise NotImplementedError("registration_client_uri was not provided.")

		r = requests.delete(self.registration_client_uri, auth=self.registration_auth)
		r.raise_for_status()



from ._oauth2 import OAuth2Client
from ._util import clean
import requests

class Client(OAuth2Client):
	def get_configuration(self):
		# Check if the URL is correct
		requests.get(self.url).raise_for_status()
		return {
			"issuer": self.url,
			"authorization_endpoint": self.url + "oauth/authorize",
			"token_endpoint": self.url + "oauth/token",
			"userinfo_endpoint": self.url + "api/v3/user",
			"response_types_supported": ["code"],
			"subject_types_supported": ["public"],
			"token_endpoint_auth_methods_supported": ["client_secret_post", "client_secret_basic"]
		}

	def translate_scope_in(self, scope):
		for scope_token in scope:
			if scope_token in ["openid", "profile", "email", "phone", "address"]:
				yield "read_user"
			else:
				yield scope_token

	def translate_scope_out(self, scope):
		for scope_token in scope:
			if scope_token == "read_user":
				yield "openid"
				yield "profile"
				yield "email"
				# GitLab does not support phone at all.
				yield "address"
			else:
				yield scope_token

	def translate_userinfo(self, userinfo):
		if userinfo["state"] != "active":
			return None

		return clean({
			"sub": userinfo.get("id"),
			"name": userinfo.get("name"),
			"nickname": userinfo.get("username"),
			"profile": userinfo.get("web_url"),
			"picture": userinfo.get("avatar_url"),
			"website": userinfo.get("website_url"),
			"email": userinfo.get("email"),
			"email_verified": True if userinfo.get("confirmed_at") else False,
			"address": {
				"formatted": userinfo.get("location"),
			},

			"openid-connect-python/bio": userinfo.get("bio"),
			"openid-connect-python/created_at": userinfo.get("created_at"),
			"openid-connect-python/is_admin": userinfo.get("is_admin"),
			"openid-connect-python/two_factor_enabled": userinfo.get("two_factor_enabled"),
			"openid-connect-python/organization": userinfo.get("organization"),

			"openid-connect-python/gitlab": {
				"can_create_group": userinfo.get("can_create_group"),
				"can_create_project": userinfo.get("can_create_project"),
				"color_scheme_id": userinfo.get("color_scheme_id"),
				"projects_limit": userinfo.get("projects_limit"),
				"theme_id": userinfo.get("theme_id"),
			}
		})

		"""
		TODO map:

		external: false
		identities:
		- extern_uid: '9737'
		provider: github
		linkedin: ''
		skype: ''
		twitter: ''
		"""



from six.moves.urllib_parse import quote


def make_netloc(host, port=None, username=None, password=None):
	netloc = ''

	if username:
		netloc = quote(username)
	if password:
		netloc += ':' + quote(password)

	if netloc:
		netloc += '@'

	if ':' in host:
		netloc += '[' + host + ']'  # IPv6 literal
	else:
		netloc += host
	if port:
		netloc += ':' + str(port)

	return netloc



from ._oauth2 import OAuth2Client
from ._util import clean
import requests

class Client(OAuth2Client):
	def __init__(self, url, client_id, client_secret, api_url=None):
		if not api_url:
			api_url = url.replace("://", "://api.")

		self._api_url = api_url
		super().__init__(url, client_id, client_secret)

	def get_configuration(self):
		# Check if the URL is correct
		requests.get(self.url).raise_for_status()
		return {
			"issuer": self.url,
			"authorization_endpoint": self.url + "login/oauth/authorize",
			"token_endpoint": self.url + "login/oauth/access_token",
			"userinfo_endpoint": self._api_url + "user",
			"response_types_supported": ["code"],
			"subject_types_supported": ["public"],
			"token_endpoint_auth_methods_supported": ["client_secret_post", "client_secret_basic"]
		}

	def translate_scope_in(self, scope):
		for scope_token in scope:
			if scope_token in ["openid", "profile", "email", "phone", "address"]:
				pass
			else:
				yield scope_token

	def translate_scope_out(self, scope):
		yield "openid"
		yield "profile"
		yield "email"
		# GitHub does not support phone at all.
		yield "address"
		for scope_token in scope:
			yield scope_token

	def translate_userinfo(self, userinfo):
		return clean({
			"sub": userinfo.get("id"),
			"name": userinfo.get("name"),
			"nickname": userinfo.get("login"),
			"profile": userinfo.get("html_url"),
			"picture": userinfo.get("avatar_url"),
			"website": userinfo.get("blog"),
			"email": userinfo.get("email"),
			"email_verified": True,
			"address": {
				"formatted": userinfo.get("location"),
			},
			"updated_at": userinfo.get("updated_at"),

			"openid-connect-python/bio": userinfo.get("bio"),
			"openid-connect-python/created_at": userinfo.get("created_at"),
			"openid-connect-python/is_admin": userinfo.get("site_admin"),
			"openid-connect-python/hireable": userinfo.get("hireable"),
			"openid-connect-python/organization": userinfo.get("company"),

			"openid-connect-python/github": {
				"events_url": userinfo.get("events_url"),
				"followers": userinfo.get("followers"),
				"followers_url": userinfo.get("followers_url"),
				"following": userinfo.get("following"),
				"following_url": userinfo.get("following_url"),
				"gists_url": userinfo.get("gists_url"),
				"organizations_url": userinfo.get("organizations_url"),
				"public_gists": userinfo.get("public_gists"),
				"public_repos": userinfo.get("public_repos"),
				"received_events_url": userinfo.get("received_events_url"),
				"repos_url": userinfo.get("repos_url"),
				"starred_url": userinfo.get("starred_url"),
				"subscriptions_url": userinfo.get("subscriptions_url"),
				"url": userinfo.get("url"),
			}
		})

		""" TODO map:

		type: User
		"""



from .. import OpenIDClient

class OAuth2Client(OpenIDClient):
	def get_id(self, token_response):
		token_response._userinfo = self.get_userinfo(token_response.access_token)
		return dict(
			iss=self.issuer,
			**token_response.userinfo,
		)



from importlib import import_module

from six.moves.urllib_parse import unquote, urlsplit, urlunsplit, parse_qs

from ._netloc import make_netloc
from ._oidc import OpenIDClient


def connect_url(url):
	components = urlsplit(url)

	try:
		protocol, scheme = components.scheme.split('+', 1)
	except ValueError:
		protocol = None
	else:
		components = components._replace(scheme=scheme)

	client_id = unquote(components.username) if components.username else None
	client_secret = unquote(components.password) if components.password else None
	components = components._replace(netloc=make_netloc(components.hostname, components.port))

	if not components.path:
		components = components._replace(path='/')

	kwargs = {}
	if components.query:
		kwargs = dict(parse_qs(components.query))
		components = components._replace(query='')

	server = urlunsplit(components)
	return connect(server, client_id, client_secret, protocol, **kwargs)

def connect(server, client_id, client_secret, protocol=None, **kwargs):
	if not protocol:
		cls = OpenIDClient
	else:
		cls = import_module('openid_connect.legacy.' + protocol).Client

	return cls(server, client_id, client_secret, **kwargs)



from requests.auth import HTTPBasicAuth


class HTTPBearerAuth(object):
	def __init__(self, token):
		self.token = token

	def __eq__(self, other):
		return self.token == getattr(other, 'token', None)

	def __ne__(self, other):
		return not self == other

	def __call__(self, r):
		r.headers['Authorization'] = 'Bearer {}'.format(self.token)
		return r




class overwritable_property(object):
	def __init__(self, func):
		self.__doc__ = getattr(func, '__doc__')
		self.func = func

	def __get__(self, obj, cls):
		if obj is None:
			return self
		return self.func(obj)
